#!/usr/bin/env python3
"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.


A simple status display for the Integration Daemon.


"""
import os
import time
from typing import Union

import nest_asyncio
from rich import box
from rich.table import Table
from textual.widgets import DataTable

from pyegeria import EgeriaTech
from pyegeria.core._exceptions import (
    PyegeriaInvalidParameterException,
    PyegeriaAPIException as PropertyServerException,
    PyegeriaUnauthorizedException as UserNotAuthorizedException,
    print_basic_exception as print_exception_response,
)

EGERIA_METADATA_STORE = os.environ.get("EGERIA_METADATA_STORE", "qs-metadata-store")
EGERIA_KAFKA_ENDPOINT = os.environ.get("KAFKA_ENDPOINT", "localhost:9092")
EGERIA_PLATFORM_URL = os.environ.get("EGERIA_PLATFORM_URL", "https://localhost:9443")
EGERIA_VIEW_SERVER = os.environ.get("EGERIA_VIEW_SERVER", "qs-view-server")
EGERIA_VIEW_SERVER_URL = os.environ.get(
    "EGERIA_VIEW_SERVER_URL", "https://localhost:9443"
)
EGERIA_INTEGRATION_DAEMON = os.environ.get("EGERIA_INTEGRATION_DAEMON", "qs-integration-daemon")
EGERIA_INTEGRATION_DAEMON_URL = os.environ.get(
    "EGERIA_INTEGRATION_DAEMON_URL", "https://localhost:9443"
)
EGERIA_USER = os.environ.get("EGERIA_USER", "erinoverview")
EGERIA_USER_PASSWORD = os.environ.get("EGERIA_USER_PASSWORD", "secret")
EGERIA_JUPYTER = bool(os.environ.get("EGERIA_JUPYTER", "False"))
EGERIA_WIDTH = int(os.environ.get("EGERIA_WIDTH", "200"))

disable_ssl_warnings = True


def add_row(
    table: Union[Table, DataTable],
    connector_name: str,
    connector_status: str,
    last_refresh_time: str,
    refresh_interval: str,
    targets_out,
    exception_msg,
) -> Table | DataTable:
    table.add_row(
        connector_name,
        connector_status,
        last_refresh_time,
        refresh_interval,
        targets_out,
        exception_msg,
    )
    return table


def display_integration_daemon_status(
    integ_server: str = EGERIA_INTEGRATION_DAEMON,
    integ_url: str = EGERIA_INTEGRATION_DAEMON_URL,
    view_server: str = EGERIA_VIEW_SERVER,
    view_url: str = EGERIA_VIEW_SERVER_URL,
    user: str = EGERIA_USER,
    user_pass: str = EGERIA_USER_PASSWORD,
    paging: bool = True,
    jupyter: bool = EGERIA_JUPYTER,
    width: int = EGERIA_WIDTH,
    sort: bool = True,
    data_table: bool = False,
) -> Table | DataTable:
    """Returns a table representing the status of connectors running on the specified Integration Daemon OMAG Server.
    This routine is meant to be used in a python script or Jupyter Notebook where the resulting table is rendered
    in either a Rich Console or a Textual Application.

    Parameters
    ----------
    integ_server : str, optional
        The name of the integration daemon server. Default is EGERIA_INTEGRATION_DAEMON.

    integ_url : str, optional
        The URL of the integration daemon server. Default is EGERIA_INTEGRATION_DAEMON_URL.

    view_server : str, optional
        The name of the view server. Default is EGERIA_VIEW_SERVER.

    view_url : str, optional
        The URL of the view server. Default is EGERIA_VIEW_SERVER_URL.

    user : str, optional
        The username for authentication. Default is EGERIA_USER.

    user_pass : str, optional
        The password for authenticated access. Default is EGERIA_USER_PASSWORD.

    paging : bool, optional
        Determines whether to use paging or a live monitor for console output. Default is True.

    jupyter : bool, optional
        Determines whether running in a Jupyter environment. Default is EGERIA_JUPYTER.

    width : int, optional
        The width of the console display. Default is EGERIA_WIDTH.

    sort : bool, optional
        Determines whether to sort the connector reports. Default is True.
    data_table: bool, optional, default is False.
        If True, a Textual DataTable widget is returned. If false, a Rich table is returned..

    Returns
    -------

    Either a Rich Table or a Textual DataTable depending on the status of data_table
    """

    s_client = EgeriaTech(view_server, view_url, user, user_pass)
    nest_asyncio.apply()

    def generate_table() -> Table:
        """Make a new table."""

        if data_table:
            table = DataTable()
            table.add_columns(
                "Connector Name",
                "Status",
                "Last Refresh Time",
                "Min Refresh (Mins)",
                "Target Element",
                "Exception Message",
            )
        else:
            table = Table(
                title=f"Integration Daemon Status @ {time.asctime()}",
                style="bold white on black",
                row_styles=["bold white on black"],
                header_style="white on dark_blue",
                title_style="bold white on black",
                caption_style="white on black",
                show_lines=True,
                box=box.ROUNDED,
                caption=f"Integration Daemon Status for Server '{integ_server}' @ Platform - {integ_url}",
                expand=True,
            )
            table.add_column("Connector Name", min_width=15, overflow="fold")
            table.add_column("Status", max_width=10, overflow="fold")
            table.add_column("Last Refresh Time", min_width=12, overflow="fold")
            table.add_column("Min Refresh (mins)", max_width=10, overflow="fold")
            table.add_column("Target Element", min_width=20, overflow="fold")
            table.add_column("Exception Message", min_width=10, overflow="fold")

        # Now get the integration connector report
        token = s_client.create_egeria_bearer_token()
        daemon_status = s_client.get_server_report(None, integ_server)

        reports = daemon_status["integrationConnectorReports"]
        if sort is True:
            connector_reports = sorted(
                reports, key=lambda x: x.get("connectorName", "---")
            )
        else:
            connector_reports = reports

        for connector in connector_reports:
            connector_name = connector.get("connectorName", "---")
            connector_status = connector.get("connectorStatus", "---")
            connector_guid = connector.get("connectorGUID", "---")
            last_refresh_time = connector.get("lastRefreshTime", "---")[:-10]
            refresh_interval = str(connector.get("minMinutesBetweenRefresh", "---"))
            exception_msg = " "
            if connector_guid != "---":
                targets = s_client.get_catalog_targets(connector_guid)
                tgt_tab = Table()
                tgt_tab.add_column("Target", overflow="fold")
                tgt_tab.add_column("UniqueName", overflow="fold")
                tgt_tab.add_column("Relationship GUID", overflow="fold")

                if type(targets) == list:
                    targets_md = True
                    for target in targets:
                        t_name = target["catalogTargetName"]
                        # t_sync = target["permittedSynchronization"]
                        t_unique_name = target["catalogTargetElement"]["uniqueName"]
                        t_rel_guid = target["relationshipGUID"]
                        # targets_m += f"* Target Name: __{t_name}__\n* Sync: {t_sync}\n* Unique Name: {t_unique_name}\n\n"
                        tgt_tab.add_row(t_name, t_unique_name, t_rel_guid)
                    # targets_md = Markdown(targets_m)
                else:
                    targets_md = False
            else:
                targets_md = False
            if targets_md is False:
                targets_out = ""
            else:
                targets_out = tgt_tab

            if connector_status in ("RUNNING", "REFRESHING", "WAITING"):
                connector_status = f"[green]{connector_status}"
            elif connector_status in ("INITIALIZE FAILED", "CONFIG_FAILED", "FAILED"):
                connector_status = f"[red]{connector_status}"
            else:
                connector_status = f"[yellow]{connector_status}"

            add_row(
                table,
                connector_name,
                connector_status,
                last_refresh_time,
                refresh_interval,
                targets_out,
                exception_msg,
            )
        return table

    try:
        return generate_table()

    except (
        PyegeriaInvalidParameterException,
        PropertyServerException,
        UserNotAuthorizedException,
    ) as e:
        print_exception_response(e)

    except KeyboardInterrupt:
        pass

    finally:
        s_client.close_session()



"""
Governance Officer Processors for Dr.Egeria v2.
"""
from typing import Dict, Any, Optional, List
from loguru import logger
import json
from inflect import engine

from pyegeria import EgeriaTech, PyegeriaException
from md_processing.v2.processors import AsyncBaseCommandProcessor
from md_processing.md_processing_utils.md_processing_constants import get_command_spec
from md_processing.md_processing_utils.common_md_utils import (
    set_gov_prop_body, set_create_body, set_update_body, 
    update_element_dictionary, set_delete_request_body,
    set_peer_gov_def_request_body, set_rel_prop_body,
    set_rel_request_body, ALL_GOVERNANCE_DEFINITIONS
)
from pyegeria.core.utils import body_slimmer

class GovernanceProcessor(AsyncBaseCommandProcessor):
    """
    Processor for Governance Definitions.
    """

    async def fetch_element(self, guid: str) -> Optional[Dict[str, Any]]:
        try:
            return await self.client._async_get_governance_definition_by_guid(guid)
        except PyegeriaException:
            return None

    async def apply_changes(self) -> str:
        verb = self.command.verb
        object_type = self.command.object_type
        attributes = self.parsed_output["attributes"]
        qualified_name = self.parsed_output["qualified_name"]
        display_name = attributes.get('Display Name', {}).get('value', qualified_name)
        
        # 1. Properties
        prop_body = set_gov_prop_body(object_type, qualified_name, attributes)

        if verb == "Update":
            guid = self.parsed_output.get("guid") or (self.as_is_element['elementHeader']['guid'] if self.as_is_element else None)
            if not guid:
                return self.command.original_text
            self.parsed_output["guid"] = guid

            body = {
                "class": "UpdateElementRequestBody",
                "properties": self.filter_update_properties(prop_body, attributes.get('Merge Update', {}).get('value', True)),
                "mergeUpdate": attributes.get('Merge Update', {}).get('value', True)
            }
            await self.client._async_update_governance_definition(guid, body)
            
            # Relationships
            await self._sync_rels(guid, attributes)
            
            logger.success(f"Updated {object_type} '{display_name}' with GUID {guid}")
            update_element_dictionary(qualified_name, {'guid': guid, 'display_name': display_name})
            return await self.render_result_markdown(guid)

        elif verb == "Create":
            body = set_create_body(object_type, attributes)
            body['properties'] = prop_body
            
            guid = await self.client._async_create_governance_definition(body_slimmer(body))
            if guid:
                self.parsed_output["guid"] = guid
                await self._sync_rels(guid, attributes)
                update_element_dictionary(qualified_name, {'guid': guid, 'display_name': display_name})
                logger.success(f"Created {object_type} '{display_name}' with GUID {guid}")
                return await self.render_result_markdown(guid)

        return self.command.original_text

    async def _sync_rels(self, guid: str, attributes: Dict[str, Any]):
        to_be_supports = attributes.get("Supports Policies", {}).get("guid_list", [])
        to_be_drivers = attributes.get("Governance Drivers", {}).get("guid_list", [])
        
        for policy in to_be_supports:
            try:
                await self.client._async_attach_supporting_definitions(policy, "GovernanceImplementation", guid)
                self.add_related_result("Supports Policy", policy)
            except Exception as e:
                self.add_related_result("Supports Policy", policy, status="failure", message=str(e))

        for driver in to_be_drivers:
            try:
                await self.client._async_attach_supporting_definitions(driver, "GovernanceResponse", guid)
                self.add_related_result("Governance Driver", driver)
            except Exception as e:
                self.add_related_result("Governance Driver", driver, status="failure", message=str(e))

class GovernanceLinkProcessor(AsyncBaseCommandProcessor):
    """
    Processor for Governance Peer/Supporting/Governed-By links.
    """



    async def fetch_as_is(self) -> Optional[Dict[str, Any]]:
        return None

    async def apply_changes(self) -> str:
        verb = self.command.verb
        object_type = self.command.object_type
        attributes = self.parsed_output["attributes"]
        
        guid1 = (attributes.get('Definition 1') or attributes.get('Referenceable', {})).get('guid')
        guid2 = (attributes.get('Definition 2') or attributes.get('Governance Definition', {})).get('guid')
        label = attributes.get('Link Label', {}).get('value', "")
        
        if not (guid1 and guid2):
            return self.command.original_text

        # Determine relationship type name
        rel_type = ""
        if "Peer" in object_type:
            parts = object_type.split()
            word = parts[2] if len(parts) >= 3 else parts[-1]
            inflector = engine()
            singular = inflector.singular_noun(word) or word
            rel_type = f"Governance{singular}Link"
        elif "Supporting" in object_type:
            rel_type = object_type.replace(' ', '')
        elif "Governed By" in object_type:
            rel_type = "GovernedBy" # Standard for Referenceable -> GovDef

        if verb in ["Link", "Attach", "Add"]:
            if "Peer" in object_type or "Supporting" in object_type:
                body = set_peer_gov_def_request_body(object_type, attributes)
                if "Supporting" in object_type:
                    body['properties'] = {
                        "class": "SupportingDefinitionProperties",
                        "typeName": rel_type,
                        "rationale": attributes.get('Rationale', {}).get('value'),
                        "effectiveFrom": attributes.get('Effective From', {}).get('value'),
                        "effectiveTo": attributes.get('Effective To', {}).get('value'),
                    }
                if "Peer" in object_type:
                    await self.client._async_link_peer_definitions(guid1, rel_type, guid2, body)
                else:
                    await self.client._async_attach_supporting_definitions(guid1, rel_type, guid2, body)
            elif "Governed By" in object_type:
                body = body_slimmer({
                    "class": "NewRelationshipRequestBody",
                    "properties": set_rel_prop_body(object_type, attributes)
                })
                await self.client._async_attach_governed_by_definition(guid1, guid2, body)
                
            logger.success(f"Linked Governance {object_type}")
            return f"\n\n# {verb} {object_type}\n\nOperation completed."

        elif verb in ["Detach", "Unlink", "Remove"]:
            body = set_delete_request_body(object_type, attributes)
            if "Peer" in object_type:
                await self.client._async_detach_peer_definitions(guid1, rel_type, guid2, body)
            elif "Supporting" in object_type:
                await self.client._async_detach_supporting_definitions(guid1, rel_type, guid2, body)
            elif "Governed By" in object_type:
                await self.client._async_detach_governed_by_definitions(guid1, guid2, body)
                
            logger.success(f"Detached Governance {object_type}")
            return f"\n\n# {verb} {object_type}\n\nOperation completed."

        return self.command.original_text

class GovernanceContextProcessor(AsyncBaseCommandProcessor):
    """
    Processor for Governance Definition Context retrieval.
    """



    async def fetch_as_is(self) -> Optional[Dict[str, Any]]:
        return None

    async def apply_changes(self) -> str:
        attributes = self.parsed_output["attributes"]
        guid = self.parsed_output.get("guid")
        output_format = attributes.get('Output Format', {}).get('value', 'LIST')
        
        if not guid:
            return self.command.original_text
            
        body = set_update_body(self.command.object_type, attributes)
        struct = await self.client._async_get_gov_def_in_context(guid, body=body, output_format=output_format)
        
        if output_format.upper() == "DICT":
            return f"\n# Context Graph for {guid}\n\n```json\n{json.dumps(struct, indent=4)}\n```"
        return f"\n# Context Graph for {guid}\n\n{struct}"



"""
PDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

Scenario tests for SchemaMaker.
"""
import time
from datetime import datetime
from typing import List, Optional
from dataclasses import dataclass

from rich import print as rprint
from rich.console import Console
from rich.table import Table

from pyegeria.omvs.schema_maker import SchemaMaker
from pyegeria.core._exceptions import PyegeriaException, PyegeriaTimeoutException

VIEW_SERVER = "qs-view-server"
PLATFORM_URL = "https://localhost:9443"
USER_ID = "peterprofile"
USER_PWD = "secret"

console = Console(width=200)


@dataclass
class TestResult:
    scenario_name: str
    status: str
    duration: float
    message: str = ""


class SchemaMakerScenarioTester:
    def __init__(self):
        self.client: Optional[SchemaMaker] = None
        self.results: List[TestResult] = []

    def setup(self) -> bool:
        try:
            self.client = SchemaMaker(VIEW_SERVER, PLATFORM_URL, USER_ID, USER_PWD)
            self.client.create_egeria_bearer_token(USER_ID, USER_PWD)
            return True
        except Exception:
            return False

    def teardown(self):
        if self.client:
            self.client.close_session()

    def scenario_schema_type_lifecycle(self) -> TestResult:
        name = "Schema Type Lifecycle"
        start_time = time.perf_counter()
        try:
            # 1. Create a schema type
            qualified_name = f"SchemaType::Scenario::{datetime.now().timestamp()}"
            body = {
                "class": "NewElementRequestBody",
                "properties": {
                    "class": "SchemaTypeProperties",
                    "qualifiedName": qualified_name,
                    "displayName": "Scenario Schema Type",
                }
            }
            guid = self.client.create_schema_type(body)
            
            # 2. Update it
            update_body = {
                "class": "UpdateElementRequestBody",
                "properties": {
                    "class": "SchemaTypeProperties",
                    "displayName": "Updated Scenario Schema Type",
                }
            }
            self.client.update_schema_type(guid, update_body)
            
            # 3. Find it
            found = self.client.find_schema_types(search_string=qualified_name)
            
            return TestResult(name, "PASSED", time.perf_counter() - start_time, f"Created, updated and found schema type {guid}")
        except PyegeriaTimeoutException as e:
            rprint(f"[bold yellow]Timeout in {name}; continuing.[/bold yellow]")
            return TestResult(name, "WARNING", time.perf_counter() - start_time, f"Timeout: {e}")
        except PyegeriaException as e:
            return TestResult(name, "FAILED", time.perf_counter() - start_time, str(e))
        except Exception as e:
            return TestResult(name, "FAILED", time.perf_counter() - start_time, f"Unexpected error: {e}")

    def run_all_scenarios(self):
        if not self.setup():
            rprint("[bold red]Setup failed, skipping scenarios[/bold red]")
            return
        
        self.results.append(self.scenario_schema_type_lifecycle())
        self.teardown()
        self.print_results_summary()

    def print_results_summary(self):
        table = Table(title="Schema Maker Scenario Results")
        table.add_column("Scenario", style="cyan")
        table.add_column("Status", justify="center")
        table.add_column("Duration (s)", justify="right")
        table.add_column("Message")
        for res in self.results:
            if res.status == "PASSED":
                style = "green"
            elif res.status == "WARNING":
                style = "yellow"
            else:
                style = "red"
            table.add_row(res.scenario_name, f"[{style}]{res.status}[/{style}]", f"{res.duration:.2f}", res.message)
        console.print(table)


def test_schema_maker_scenarios():
    tester = SchemaMakerScenarioTester()
    tester.run_all_scenarios()


if __name__ == "__main__":
    test_schema_maker_scenarios()



from __future__ import annotations

import json
import os
from typing import Dict, Any, Iterable

from .parse_compact_export import load_compact_json, expand_command


def _to_semicolon_str(value: Any) -> str:
    if value is None:
        return ""
    if isinstance(value, str):
        return value
    if isinstance(value, (list, tuple, set)):
        return "; ".join([str(v) for v in value if str(v).strip()])
    return str(value)


def _command_to_spec(name: str, cdef: Dict[str, Any], expanded: Dict[str, Any]) -> Dict[str, Any]:
    # Copy known top-level fields if present in the compact command
    spec: Dict[str, Any] = {}
    for key in (
        "display_name",
        "qn_prefix",
        "family",
        "description",
        "verb",
        "upsert",
        "attach",
        "ReferenceURL",
        "level",
        "find_method",
        "find_constraints",
        "extra_find",
        "extra_constraints",
        "Journal Entry",
    ):
        if key in cdef:
            spec[key] = cdef[key]

    # Normalize alternate_names to the legacy semicolon string
    alt = cdef.get("alternate_names", "")
    spec["alternate_names"] = _to_semicolon_str(alt)

    # Use expanded attributes directly (already resolved and expanded)
    # We ensure each attribute has a 'style' for processor compatibility
    attrs = expanded.get("all_attributes", [])
    for attr in attrs:
        if not attr.get("style"):
            attr["style"] = "Simple"
            
    spec["Attributes"] = attrs

    return spec


def load_compact_specs_from_dir(dir_path: str, families_allowlist: Iterable[str] | None = None) -> Dict[str, Dict[str, Any]]:
    """
    Load all compact command files from a directory and return a mapping 
    for COMMAND_DEFINITIONS["Command Specifications"].

    Args:
        dir_path: Path to directory with compact JSON files.
        families_allowlist: If provided, include only commands whose 'family' is in this set.

    Returns:
        Dict[str, Dict]: { command_name: command_spec }
    """
    if families_allowlist:
        families = {f.strip() for f in families_allowlist if f and f.strip()}
    else:
        families = set()

    specs: Dict[str, Dict[str, Any]] = {}
    if not os.path.isdir(dir_path):
        return specs

    all_attr_defs = {}
    all_bundles = {}
    loaded_files = []

    from loguru import logger

    for fname in sorted(os.listdir(dir_path)):
        if not fname.endswith(".json"):
            continue
        fpath = os.path.join(dir_path, fname)
        try:
            data = load_compact_json(fpath)

            # Merge attribute definitions with conflict detection
            for aname, adef in data.get("attribute_definitions", {}).items():
                if aname in all_attr_defs and all_attr_defs[aname] != adef:
                    logger.debug(f"Attribute '{aname}' in {fname} differs from previous definition.")
                all_attr_defs[aname] = adef

            # Merge bundle definitions with conflict detection
            for bname, bdef in data.get("bundles", {}).items():
                if bname in all_bundles and all_bundles[bname] != bdef:
                    logger.debug(f"Bundle '{bname}' in {fname} differs from previous definition.")
                all_bundles[bname] = bdef

            loaded_files.append((fname, data))
        except Exception as e:
            logger.warning(f"Failed to load compact JSON {fname}: {e}")
            continue

    for fname, data in loaded_files:
        file_commands = data.get("commands", {})

        for cname, cdef in file_commands.items():
            fam = cdef.get("family")
            if families and fam not in families:
                continue

            if cname in specs:
                logger.warning(f"Command '{cname}' in {fname} overwrites a definition from a previous file.")

            try:
                expanded = expand_command(cdef, all_bundles, all_attr_defs)
                specs[cname] = _command_to_spec(cname, cdef, expanded)
            except (KeyError, ValueError) as e:
                logger.warning(f"Skipping command '{cname}' from {fname}: {e}")
                continue

    return specs



#!/usr/bin/env python3
"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.


A simple display for glossary terms
"""
import argparse
import os
import sys
import time

from rich import box
from rich.console import Console
from rich.markdown import Markdown
from rich.prompt import Prompt
from rich.table import Table
from rich.text import Text

from pyegeria import (
    EgeriaTech, settings,
    PyegeriaAPIException, PyegeriaClientException, print_basic_exception, print_exception_table
)
from pyegeria.core._globals import NO_GLOSSARIES_FOUND
app_config = settings.Environment
# config_logging()
disable_ssl_warnings = True


EGERIA_USER = os.environ.get("EGERIA_USER", "erinoverview")
EGERIA_USER_PASSWORD = os.environ.get("EGERIA_USER_PASSWORD", "secret")


def display_glossary_terms(
    search_string: str = "*",
    glossary_guid: str = settings.User_Profile.egeria_home_glossary_name,
    glossary_name: str = None,
    view_server: str = app_config.egeria_view_server,
    view_url: str = app_config.egeria_view_server_url,
    user_id: str = EGERIA_USER,
    user_pass: str = EGERIA_USER_PASSWORD,
    jupyter: bool = app_config.egeria_jupyter,
    width: int = app_config.console_width,
    output_format: str = "TABLE",
):
    """Display a table of glossary terms filtered by search_string and glossary, if specified. If no
        filters then all terms are displayed. If glossary_guid or name is specified, then only terms from that
        glossary are displayed.
    Parameters
    ----------
    search_string : str, optional
        The string to search for terms. Defaults to "*".
    glossary_guid : str, optional
        The unique identifier of the glossary. Defaults to None. If specified, then only terms from that glossary
        are displayed. If both glossary_guid and glossary_name are provided then glossary_guid will take precedence.
    glossary_name : str, optional
        The display name of the glossary. Defaults to None. If specified, then only terms from that glossary
        are displayed. If both glossary_guid and glossary_name are provided then glossary_guid will take precedence.
        Note that the use of glossary display name relies on the qualified name conforming to convention. GUID is more
        reliable.
    view_server : str
        The server where the glossary is hosted. Defaults to EGERIA_VIEW_SERVER.
    view_url : str
        The URL of the server where the glossary is hosted. Defaults to EGERIA_VIEW_SERVER_URL.
    user_id : str
        The user ID for authentication. Defaults to EGERIA_USER.
    user_pass : str
        The user password for authentication. Defaults to EGERIA_USER_PASSWORD.
    jupyter : bool
        Flag to indicate if the output should be formatted for Jupyter notebook. Defaults to EGERIA_JUPYTER.
    width : int
        The width of the console output. Defaults to EGERIA_WIDTH.
        output_format: str, optional, default is 'JSON'
        One of TABLE, FORM, REPORT
    """

    console = Console(
        style="bold bright_white on black", width=width, force_terminal=not jupyter
    )
    try:
        g_client = EgeriaTech(view_server, view_url, user_id, user_pass)
        token = g_client.create_egeria_bearer_token(user_id, user_pass)
        if (glossary_name is not None) and (glossary_name != "*"):
            glossary_guid = g_client.get_guid_for_name(glossary_name)
            if glossary_guid == NO_GLOSSARIES_FOUND:
                console.print(
                    f"\nThe glossary name {glossary_name} was not found. Please try using the glossary guid"
                )
                sys.exit(1)
        elif (glossary_guid is not None) and (len(glossary_guid) < 10):
                glossary_guid = None

        if output_format == "FORM":
            action = "Update-Form"
        elif output_format == "REPORT":
            action = "Report"
        if output_format != "TABLE":
            file_path = os.path.join(app_config.pyegeria_root, app_config.egeria_outbox)
            file_name = f"Terms-{time.strftime('%Y-%m-%d-%H-%M-%S')}-{action}.md"
            full_file_path = os.path.join(file_path, file_name)
            os.makedirs(os.path.dirname(full_file_path), exist_ok=True)
            output = g_client.find_glossary_terms(search_string, glossary_guid, output_format=output_format)
            if output == "NO_TERMS_FOUND":
                print(f"\n==> No terms found for search string '{search_string}'")
                return
            with open(full_file_path, 'w') as f:
                f.write(output)
            print(f"\n==> Terms output written to {full_file_path}")
            return

    except (PyegeriaAPIException, PyegeriaClientException) as e:
        print_basic_exception(e)



    def generate_table(search_string: str, glossary_guid: str) -> Table:
        """Make a new table."""
        table = Table(
            title=f"Glossary Definitions for Terms like  {search_string} @ {time.asctime()}",
            style="bright_white on black",
            # row_styles="bright_white on black",
            header_style="bright_white on dark_blue",
            title_style="bold white on black",
            caption_style="white on black",
            show_lines=True,
            box=box.ROUNDED,
            caption=f"View Server '{view_server}' @ Platform - {view_url}",
            expand=True,
        )
        table.add_column("Term Name / Abbreviation / Version")
        table.add_column("Qualified Name / GUID / Aliases", width=38, no_wrap=True)
        table.add_column("Summary")
        table.add_column("Description")
        table.add_column("Glossary")
        table.add_column("Collection Folder")
        table.add_column("Status")
        table.add_column("Example/Usage", min_width=50)

        terms = g_client.find_glossary_terms(
            search_string,
            page_size=500,
        )

        if type(terms) is str:
            print(f"No terms found! - {search_string} : {glossary_guid} ")
            sys.exit(0)
        sorted_terms = sorted(
            terms, key=lambda k: k["properties"].get("displayName","---")
        )
        style = "bright_white on black"
        if type(terms) is str:
            return table
        glossary_info = {}
        for term in sorted_terms:
            props = term.get("properties", "None")
            if props == "None":
                return table

            display_name = props.get("displayName","---")
            qualified_name = props["qualifiedName"]
            term_guid = term["elementHeader"]["guid"]
            aliases = props.get("aliases", "---")
            q_name = Text(
                f"{qualified_name}\n&\n{term_guid}\n&\n{aliases}", style=style, justify="center"
            )
            abbrev = props.get("abbreviation", "---")
            summary = props.get("summary", "---")
            description = Markdown(props.get("description",'---'))
            version = props.get("publishVersionIdentifier", "---")
            example = props.get("example", "---")
            usage = props.get("usage", "---")
            ex_us_out = Markdown(f"Example:\n{example}\n---\nUsage: \n{usage}")

            parents = term.get('memberOfCollections',[])
            glossary_list_md = ""
            category_list_md = ""
            for parent in parents:
                parent_q_name = parent.get('relatedElement',{}).get('properties',{}).get('qualifiedName','---')
                parent_type = parent.get('relatedElement',{}).get('elementHeader',{}).get('type',{}).get('typeName','---')
                if parent_type == 'Glossary':
                    glossary_list_md += f"{parent_q_name}\n"
                elif parent_type == 'CollectionFolder':
                    category_list_md += f"{parent_q_name}\n"


            term_abb_ver_out = Markdown(f"{display_name}\n---\n{abbrev}\n---\n{version}")
            category_list_out = Markdown(category_list_md)

            term_status = term["elementHeader"].get("status","---")
            table.add_row(
                term_abb_ver_out,
                q_name,
                summary,
                description,
                glossary_list_md,
                category_list_out,
                term_status,
                ex_us_out,
                style="bold white on black",
            )

        g_client.close_session()
        return table

    try:
        with console.pager(styles=True):
            console.print(generate_table(search_string, glossary_guid))

    except (PyegeriaAPIException, PyegeriaClientException) as e:
        print_basic_exception(e)


def main():
    sus_guid = "f9b78b26-6025-43fa-9299-a905cc6d1575"
    parser = argparse.ArgumentParser()
    parser.add_argument("--server", help="Name of the server to display status for")
    parser.add_argument("--url", help="URL Platform to connect to")
    parser.add_argument("--userid", help="User Id")
    parser.add_argument("--password", help="User Password")
    parser.add_argument("--guid", help="GUID of glossary to search")

    args = parser.parse_args()

    # server = args.server if args.server is not None else EGERIA_VIEW_SERVER
    server = args.server if args.server is not None else app_config.egeria_view_server

    url = args.url if args.url is not None else app_config.egeria_view_server_url
    userid = args.userid if args.userid is not None else EGERIA_USER
    user_pass = args.password if args.password is not None else EGERIA_USER_PASSWORD
    guid = args.guid if args.guid is not None else settings.User_Profile.egeria_home_glossary_name

    try:
        search_string = Prompt.ask("Enter the term you are searching for:", default="*")
        glossary_name = Prompt.ask(
            "Enter the name of the glossary to search or '*' for all glossaries:",
            default="*",
        )
        output_format = Prompt.ask("What output format do you want?", choices=["TABLE", "FORM", "REPORT"], default="TABLE")

        display_glossary_terms(search_string, guid, glossary_name, server, url, userid, user_pass,
                               output_format=output_format)

    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

This module provides access to the Product Manager OMVS module.

The Product Manager OMVS provides APIs for managing digital products and their
relationships, including product dependencies and product managers.
"""

import asyncio
from typing import Optional

from pyegeria.omvs.collection_manager import CollectionManager
from pyegeria.core._server_client import ServerClient
from pyegeria.models import (
    NewElementRequestBody,
    UpdateElementRequestBody,
    NewRelationshipRequestBody,
    DeleteRelationshipRequestBody,
    DeleteElementRequestBody,
)
from pyegeria.core.utils import dynamic_catch, body_slimmer
from loguru import logger


class ProductManager(CollectionManager):
    """
    Manage digital products, digital product catalogs, and their relationships.

    This client provides methods to create, update, and manage digital products and
    digital product catalogs, including linking product dependencies and product manager roles.

    Attributes
    ----------
    view_server : str
        The name of the View Server to connect to.
    platform_url : str
        URL of the server platform to connect to.
    user_id : str
        The identity of the user calling the method - this sets a default optionally
        used by the methods when the user doesn't pass the user_id on a method call.
    user_pwd : str, optional
        The password associated with the user_id. Defaults to None.
    token : str, optional
        An optional bearer token for authentication.

    Methods
    -------
    create_digital_product(body)
        Create a new digital product collection.
    update_digital_product(digital_product_guid, body)
        Update the properties of a digital product.
    delete_digital_product(digital_product_guid, body, cascade)
        Delete a digital product.
    get_digital_product_by_guid(digital_product_guid, body, output_format, report_spec)
        Return the properties of a specific digital product by GUID.
    get_digital_products_by_name(filter_string, body, start_from, page_size, output_format, report_spec)
        Returns the list of digital products with a particular name.
    find_digital_products(search_string, starts_with, ends_with, ignore_case, start_from, page_size, output_format, report_spec, body)
        Returns the list of digital products matching the search string.
    create_digital_product_catalog(body)
        Create a new digital product catalog collection.
    update_digital_product_catalog(digital_product_catalog_guid, body)
        Update the properties of a digital product catalog.
    delete_digital_product_catalog(digital_product_catalog_guid, body, cascade)
        Delete a digital product catalog.
    get_digital_product_catalog_by_guid(digital_product_catalog_guid, body, output_format, report_spec)
        Return the properties of a specific digital product catalog by GUID.
    get_digital_product_catalogs_by_name(filter_string, body, start_from, page_size, output_format, report_spec)
        Returns the list of digital product catalogs with a particular name.
    find_digital_product_catalogs(search_string, starts_with, ends_with, ignore_case, start_from, page_size, output_format, report_spec, body)
        Returns the list of digital product catalogs matching the search string.
    link_digital_product_dependency(consumer_product_guid, consumed_product_guid, body)
        Link two dependent digital products.
    detach_digital_product_dependency(consumer_product_guid, consumed_product_guid, body)
        Unlink dependent digital products.
    link_product_manager(digital_product_guid, product_manager_role_guid, body)
        Attach a product manager role to a digital product.
    detach_product_manager(digital_product_guid, product_manager_role_guid, body)
        Detach a product manager from a digital product.
    """

    def __init__(
        self,
        view_server: str = None,
        platform_url: str = None,
        user_id: str = None,
        user_pwd: Optional[str] = None,
        token: Optional[str] = None,
    ):
        ServerClient.__init__(self, view_server, platform_url, user_id, user_pwd, token)
        self.view_server = self.server_name
        self.platform_url = self.platform_url
        self.user_id = self.user_id
        self.user_pwd = self.user_pwd
        self.product_manager_command_root: str = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/product-manager"
        )

    def _prepare_body(self, body: Optional[dict | NewElementRequestBody | UpdateElementRequestBody | 
                                           NewRelationshipRequestBody | DeleteRelationshipRequestBody]) -> dict:
        """Convert Pydantic models to dict and slim the body."""
        if body is None:
            return {}
        if isinstance(body, dict):
            return body_slimmer(body)
        # It's a Pydantic model
        return body_slimmer(body.model_dump(mode='json', by_alias=True, exclude_none=True))

    #
    # Digital Product Management
    #

    @dynamic_catch
    async def _async_create_digital_product(
        self,
        body: Optional[dict | NewElementRequestBody] = None,
    ) -> str:
        """Create a new digital product collection. Async version.

        Parameters
        ----------
        body : dict | NewElementRequestBody, optional
            Request body containing digital product properties.

        Returns
        -------
        str
            The GUID of the created digital product.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        ValidationError
            If the body does not conform to NewElementRequestBody.
        PyegeriaNotAuthorizedException
            If the user is not authorized for the requested action.

        Notes
        -----
        Sample JSON body:
        ```json
        {
          "class" : "NewElementRequestBody",
          "isOwnAnchor" : true,
          "properties": {
            "class" : "DigitalProductProperties",
            "qualifiedName": "DigitalProduct::Product Name",
            "displayName" : "Product Display Name",
            "description" : "Description of the product",
            "identifier" : "Product ID",
            "productName" : "Product Name",
            "introductionDate" : "2024-01-01T00:00:00.000+00:00"
          }
        }
        ```
        """
        url = f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/product-manager/collections"
        return await self._async_create_element_body_request(url, ["DigitalProductProperties"], body)

    def create_digital_product(
        self,
        body: Optional[dict | NewElementRequestBody] = None,
    ) -> str:
        """Create a new digital product collection.

        Parameters
        ----------
        body : dict | NewElementRequestBody, optional
            Request body containing digital product properties.

        Returns
        -------
        str
            The GUID of the created digital product.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.

        Notes
        -----
        Sample JSON body:
        ```json
        {
          "class" : "NewElementRequestBody",
          "isOwnAnchor" : true,
          "properties": {
            "class" : "DigitalProductProperties",
            "qualifiedName": "DigitalProduct::Product Name",
            "displayName" : "Product Display Name",
            "description" : "Description of the product",
            "identifier" : "Product ID",
            "productName" : "Product Name",
            "introductionDate" : "2024-01-01T00:00:00.000+00:00"
          }
        }
        ```
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self._async_create_digital_product(body))

    @dynamic_catch
    async def _async_update_digital_product(
        self,
        digital_product_guid: str,
        body: Optional[dict | UpdateElementRequestBody] = None,
    ) -> None:
        """Update the properties of a digital product. Async version.

        Parameters
        ----------
        digital_product_guid : str
            The GUID of the digital product to update.
        body : dict | UpdateElementRequestBody, optional
            Request body containing updated properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.

        Notes
        -----
        Sample JSON body:
        ```json
        {
          "class" : "UpdateElementRequestBody",
          "mergeUpdate": true,
          "properties": {
            "class" : "DigitalProductProperties",
            "displayName" : "New Display Name"
          }
        }
        ```
        """
        url = f"{self.product_manager_command_root}/collections/{digital_product_guid}/update"
        await self._async_update_element_body_request(url, ["DigitalProductProperties"], body)

    def update_digital_product(
        self,
        digital_product_guid: str,
        body: Optional[dict | UpdateElementRequestBody] = None,
    ) -> None:
        """Update the properties of a digital product.

        Parameters
        ----------
        digital_product_guid : str
            The GUID of the digital product to update.
        body : dict | UpdateElementRequestBody, optional
            Request body containing updated properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.

        Notes
        -----
        Sample JSON body:
        ```json
        {
          "class" : "UpdateElementRequestBody",
          "mergeUpdate": true,
          "properties": {
            "class" : "DigitalProductProperties",
            "displayName" : "New Display Name"
          }
        }
        ```
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_update_digital_product(digital_product_guid, body)
        )

    @dynamic_catch
    async def _async_delete_digital_product(
        self,
        digital_product_guid: str,
        body: Optional[dict | DeleteElementRequestBody] = None,
        cascade: bool = False,
    ) -> None:
        """Delete a digital product. Async version.

        Parameters
        ----------
        digital_product_guid : str
            The GUID of the digital product to delete.
        body : dict | DeleteElementRequestBody, optional
            Request body for deletion.
        cascade : bool, optional, default=False
            If true, performs a cascade delete.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.

        Notes
        -----
        Sample JSON body:
        ```json
        {
          "class": "DeleteElementRequestBody",
          "cascadedDelete": false
        }
        ```
        """
        if body is None:
            body = {"class": "DeleteElementRequestBody"}
        url = f"{self.product_manager_command_root}/collections/{digital_product_guid}/delete"
        await self._async_delete_element_request(url, body, cascade)
        logger.info(f"Deleted digital product {digital_product_guid} with cascade {cascade}")

    def delete_digital_product(
        self,
        digital_product_guid: str,
        body: Optional[dict | DeleteElementRequestBody] = None,
        cascade: bool = False,
    ) -> None:
        """Delete a digital product.

        Parameters
        ----------
        digital_product_guid : str
            The GUID of the digital product to delete.
        body : dict | DeleteElementRequestBody, optional
            Request body for deletion.
        cascade : bool, optional, default=False
            If true, performs a cascade delete.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.

        Notes
        -----
        Sample JSON body:
        ```json
        {
          "class": "DeleteElementRequestBody",
          "cascadedDelete": false
        }
        ```
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_delete_digital_product(digital_product_guid, body, cascade)
        )

    @dynamic_catch
    async def _async_get_digital_product_by_guid(
        self,
        digital_product_guid: str,
        body: Optional[dict] = None,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
    ) -> dict | str:
        """Return the properties of a specific digital product. Async version.

        Parameters
        ----------
        digital_product_guid : str
            Unique identifier of the digital product.
        body : dict, optional
            Full request body.
        output_format : str, default="JSON"
            One of "JSON", "DICT", "MD", "FORM", "REPORT", or "MERMAID".
        report_spec : str | dict, optional
            The desired output columns/fields to include.

        Returns
        -------
        dict | str
            A JSON dict representing the specified digital product.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.

        Notes
        -----
        Body sample:
        {
          "class": "GetRequestBody",
          "asOfTime": "{{$isoTimestamp}}",
          "effectiveTime": "{{$isoTimestamp}}",
          "forLineage": false,
          "forDuplicateProcessing": false
        }
        """
        url = f"{self.product_manager_command_root}/collections/{digital_product_guid}/retrieve"
        response = await self._async_get_guid_request(
            url,
            _type="DigitalProduct",
            _gen_output=self._generate_collection_output,
            output_format=output_format,
            report_spec=report_spec,
            body=body,
        )
        return response

    def get_digital_product_by_guid(
        self,
        digital_product_guid: str,
        body: Optional[dict] = None,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
    ) -> dict | str:
        """Return the properties of a specific digital product. Sync version.

        Parameters
        ----------
        digital_product_guid : str
            Unique identifier of the digital product.
        body : dict, optional
            Full request body.
        output_format : str, default="JSON"
            One of "JSON", "DICT", "MD", "FORM", "REPORT", or "MERMAID".
        report_spec : str | dict, optional
            The desired output columns/fields to include.

        Returns
        -------
        dict | str
            A JSON dict representing the specified digital product.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self._async_get_digital_product_by_guid(
                digital_product_guid, body, output_format, report_spec
            )
        )

    @dynamic_catch
    async def _async_get_digital_products_by_name(
        self,
        filter_string: str,
        classification_names: Optional[list[str]] = None,
        body: Optional[dict] = None,
        start_from: int = 0,
        page_size: int = 0,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
    ) -> list | str:
        """Returns the list of digital products with a particular name. Async version.

        Parameters
        ----------
        filter_string : str
            Name to use to find matching digital products.
        classification_names : list[str], optional
            List of classification names to filter on.
        body : dict, optional
            Provides a full request body. If specified, supersedes the filter_string parameter.
        start_from : int, default=0
            When multiple pages of results are available, the page number to start from.
        page_size : int, default=0
            The number of items to return in a single page.
        output_format : str, default="JSON"
            One of "JSON", "DICT", "MD", "FORM", "REPORT", or "MERMAID".
        report_spec : str | dict, optional
            The desired output columns/fields to include.

        Returns
        -------
        list | str
            A list of digital products matching the name.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        url = f"{self.product_manager_command_root}/collections/by-name"
        response = await self._async_get_name_request(url, _type="DigitalProduct",
                                                      _gen_output=self._generate_collection_output,
                                                      filter_string=filter_string,
                                                      classification_names=classification_names, start_from=start_from,
                                                      page_size=page_size, output_format=output_format,
                                                      report_spec=report_spec, body=body)
        return response

    def get_digital_products_by_name(
        self,
        filter_string: str,
        classification_names: Optional[list[str]] = None,
        body: Optional[dict] = None,
        start_from: int = 0,
        page_size: int = 0,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
    ) -> list | str:
        """Returns the list of digital products with a particular name. Sync version.

        Parameters
        ----------
        filter_string : str
            Name to use to find matching digital products.
        classification_names : list[str], optional
            List of classification names to filter on.
        body : dict, optional
            Provides a full request body. If specified, supersedes the filter_string parameter.
        start_from : int, default=0
            When multiple pages of results are available, the page number to start from.
        page_size : int, default=0
            The number of items to return in a single page.
        output_format : str, default="JSON"
            One of "JSON", "DICT", "MD", "FORM", "REPORT", or "MERMAID".
        report_spec : str | dict, optional
            The desired output columns/fields to include.

        Returns
        -------
        list | str
            A list of digital products matching the name.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self._async_get_digital_products_by_name(
                filter_string, classification_names, body, start_from, page_size, output_format, report_spec
            )
        )

    @dynamic_catch
    async def _async_find_digital_products(
        self,
        search_string: str = "*",
        body: Optional[dict] = None,
        starts_with: bool = True,
        ends_with: bool = False,
        ignore_case: bool = False,
        start_from: int = 0,
        page_size: int = 100,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
        metadata_element_subtype: Optional[list[str]] = None,
        **kwargs
    ) -> list | str:
        """Returns the list of digital products matching the search string. Async version.

        Parameters
        ----------
        search_string : str, default="*"
            Search string to match against. '*' matches all digital products.
        starts_with : bool, default=True
            Starts with the supplied string.
        ends_with : bool, default=False
            Ends with the supplied string.
        ignore_case : bool, default=False
            Ignore case when searching.
        anchor_domain : str, optional
            Anchor domain to filter on.
        metadata_element_type : str, optional
            Metadata element type name to filter on.
        metadata_element_subtype : list[str], optional
            List of metadata element subtypes to filter on.
        skip_relationships : list[str], optional
            List of relationship types to skip.
        include_only_relationships : list[str], optional
            List of relationship types to include only.
        skip_classified_elements : list[str], optional
            List of classification names to skip.
        include_only_classified_elements : list[str], optional
            List of classification names to include only.
        graph_query_depth : int, default=3
            Depth of graph query.
        governance_zone_filter : list[str], optional
            List of governance zones to filter on.
        as_of_time : str, optional
            Time for historical queries.
        effective_time : str, optional
            Effective time for the query.
        relationship_page_size : int, default=0
            Page size for relationships.
        limit_results_by_status : list[str], optional
            List of statuses to limit results by.
        sequencing_order : str, optional
            Sequencing order for results.
        sequencing_property : str, optional
            Property to sequence by.
        start_from : int, default=0
            When multiple pages of results are available, the page number to start from.
        page_size : int, default=100
            The number of items to return in a single page.
        output_format : str, default="JSON"
            One of "JSON", "DICT", "MD", "FORM", "REPORT", or "MERMAID".
        report_spec : str | dict, optional
            The desired output columns/fields to include.
        property_names: list[str], optional
            The names of properties to search for.
        body : dict, optional
            If provided, the search parameters in the body supersede other attributes.

        Returns
        -------
        list | str
            Output depends on the output format specified.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        if metadata_element_subtype is None:
            metadata_element_subtype = ["DigitalProduct"]

        url = f"{self.product_manager_command_root}/collections/by-search-string"
        
        # Merge explicit parameters with kwargs
        params = {
            'search_string': search_string,
            'body': body,
            'starts_with': starts_with,
            'ends_with': ends_with,
            'ignore_case': ignore_case,
            'start_from': start_from,
            'page_size': page_size,
            'output_format': output_format,
            'report_spec': report_spec,
            'metadata_element_subtypes': metadata_element_subtype
        }
        params.update(kwargs)
        
        # Filter out None values, but keep search_string even if None (it's required)
        params = {k: v for k, v in params.items() if v is not None or k == 'search_string'}
        
        response = await self._async_find_request(url, _type="DigitalProduct",
                                                  _gen_output=self._generate_collection_output, **params)
        return response

    def find_digital_products(
        self,
        search_string: str = "*",
        body: Optional[dict] = None,
        starts_with: bool = True,
        ends_with: bool = False,
        ignore_case: bool = False,
        start_from: int = 0,
        page_size: int = 100,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
        metadata_element_subtype: Optional[list[str]] = None,
        **kwargs
    ) -> list | str:
        """Returns the list of digital products matching the search string. Sync version.

        Parameters
        ----------
        search_string : str, default="*"
            Search string to match against. '*' matches all digital products.
        starts_with : bool, default=True
            Starts with the supplied string.
        ends_with : bool, default=False
            Ends with the supplied string.
        ignore_case : bool, default=False
            Ignore case when searching.
        anchor_domain : str, optional
            Anchor domain to filter on.
        metadata_element_type : str, optional
            Metadata element type name to filter on.
        metadata_element_subtype : list[str], optional
            List of metadata element subtypes to filter on.
        skip_relationships : list[str], optional
            List of relationship types to skip.
        include_only_relationships : list[str], optional
            List of relationship types to include only.
        skip_classified_elements : list[str], optional
            List of classification names to skip.
        include_only_classified_elements : list[str], optional
            List of classification names to include only.
        graph_query_depth : int, default=3
            Depth of graph query.
        governance_zone_filter : list[str], optional
            List of governance zones to filter on.
        as_of_time : str, optional
            Time for historical queries.
        effective_time : str, optional
            Effective time for the query.
        relationship_page_size : int, default=0
            Page size for relationships.
        limit_results_by_status : list[str], optional
            List of statuses to limit results by.
        sequencing_order : str, optional
            Sequencing order for results.
        sequencing_property : str, optional
            Property to sequence by.
        start_from : int, default=0
            When multiple pages of results are available, the page number to start from.
        page_size : int, default=100
            The number of items to return in a single page.
        output_format : str, default="JSON"
            One of "JSON", "DICT", "MD", "FORM", "REPORT", or "MERMAID".
        report_spec : str | dict, optional
            The desired output columns/fields to include.
        property_names: list[str], optional
            The names of properties to search for.
        body : dict, optional
            If provided, the search parameters in the body supersede other attributes.

        Returns
        -------
        list | str
            Output depends on the output format specified.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self._async_find_digital_products(
                search_string=search_string,
                body=body,
                starts_with=starts_with,
                ends_with=ends_with,
                ignore_case=ignore_case,
                start_from=start_from,
                page_size=page_size,
                output_format=output_format,
                report_spec=report_spec,
                metadata_element_subtype=metadata_element_subtype,
                **kwargs
            )
        )

    #
    # Digital Product Dependency Management
    #

    @dynamic_catch
    async def _async_link_digital_product_dependency(
        self,
        consumer_product_guid: str,
        consumed_product_guid: str,
        body: Optional[dict | NewRelationshipRequestBody] = None,
    ) -> None:
        """Link two dependent digital products. Async version.

        Parameters
        ----------
        consumer_product_guid : str
            The GUID of the digital product that consumes another.
        consumed_product_guid : str
            The GUID of the digital product being consumed.
        body : dict | NewRelationshipRequestBody, optional
            Request body containing relationship properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.

        Notes
        -----
        Sample JSON body:
        ```json
        {
          "class" : "NewRelationshipRequestBody",
          "properties": {
            "class": "DigitalProductDependencyProperties",
            "label": "add label here",
            "description": "add description here"
          }
        }
        ```
        """
        url = (
            f"{self.product_manager_command_root}/digital-products/"
            f"{consumer_product_guid}/product-dependencies/{consumed_product_guid}/attach"
        )
        await self._async_new_relationship_request(url, ["DigitalProductDependencyProperties"], body)
        logger.info(f"Linked {consumed_product_guid} -> {consumer_product_guid}")

    def link_digital_product_dependency(
        self,
        consumer_product_guid: str,
        consumed_product_guid: str,
        body: Optional[dict | NewRelationshipRequestBody] = None,
    ) -> None:
        """Link two dependent digital products.

        Parameters
        ----------
        consumer_product_guid : str
            The GUID of the digital product that consumes another.
        consumed_product_guid : str
            The GUID of the digital product being consumed.
        body : dict | NewRelationshipRequestBody, optional
            Request body containing relationship properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.

        Notes
        -----
        Sample JSON body:
        ```json
        {
          "class" : "NewRelationshipRequestBody",
          "properties": {
            "class": "DigitalProductDependencyProperties",
            "label": "add label here",
            "description": "add description here"
          }
        }
        ```
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_link_digital_product_dependency(
                consumer_product_guid, consumed_product_guid, body
            )
        )

    @dynamic_catch
    async def _async_detach_digital_product_dependency(
        self,
        consumer_product_guid: str,
        consumed_product_guid: str,
        body: Optional[dict | DeleteRelationshipRequestBody] = None,
    ) -> None:
        """Unlink dependent digital products. Async version.

        Parameters
        ----------
        consumer_product_guid : str
            The GUID of the consumer digital product.
        consumed_product_guid : str
            The GUID of the consumed digital product to detach.
        body : dict | DeleteRelationshipRequestBody, optional
            Request body for deletion.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.

        Notes
        -----
        Sample JSON body:
        ```json
        {
          "class" : "DeleteRelationshipRequestBody"
        }
        ```
        """
        url = (
            f"{self.product_manager_command_root}/digital-products/"
            f"{consumer_product_guid}/product-dependencies/{consumed_product_guid}/detach"
        )
        await self._async_delete_relationship_request(url, body)
        logger.info(f"Detached digital product dependency {consumer_product_guid} -> {consumed_product_guid}")

    def detach_digital_product_dependency(
        self,
        consumer_product_guid: str,
        consumed_product_guid: str,
        body: Optional[dict | DeleteRelationshipRequestBody] = None,
    ) -> None:
        """Unlink dependent digital products.

        Parameters
        ----------
        consumer_product_guid : str
            The GUID of the consumer digital product.
        consumed_product_guid : str
            The GUID of the consumed digital product to detach.
        body : dict | DeleteRelationshipRequestBody, optional
            Request body for deletion.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.

        Notes
        -----
        Sample JSON body:
        ```json
        {
          "class" : "DeleteRelationshipRequestBody"
        }
        ```
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_detach_digital_product_dependency(
                consumer_product_guid, consumed_product_guid, body
            )
        )

    #
    # Product Manager Role Management
    #

    @dynamic_catch
    async def _async_link_product_manager(
        self,
        digital_product_guid: str,
        product_manager_role_guid: str,
        body: Optional[dict | NewRelationshipRequestBody] = None,
    ) -> None:
        """Attach a product manager role to a digital product. Async version.

        Parameters
        ----------
        digital_product_guid : str
            The GUID of the digital product.
        product_manager_role_guid : str
            The GUID of the product manager role.
        body : dict | NewRelationshipRequestBody, optional
            Request body containing relationship properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.

        Notes
        -----
        Sample JSON body:
        ```json
        {
          "class": "NewRelationshipRequestBody",
          "properties": {
              "class" : "AssignmentScopeProperties",
              "assignmentType": "Product Manager",
              "description": "The person or role responsible for the product"
          }
        }
        ```
        """
        url = (
            f"{self.product_manager_command_root}/digital-products/"
            f"{digital_product_guid}/product-managers/{product_manager_role_guid}/attach"
        )
        await self._async_new_relationship_request(url, ["AssignmentScopeProperties"], body)
        logger.info(f"Attached digital product manager {digital_product_guid} -> {product_manager_role_guid}")

    def link_product_manager(
        self,
        digital_product_guid: str,
        product_manager_role_guid: str,
        body: Optional[dict | NewRelationshipRequestBody] = None,
    ) -> None:
        """Attach a product manager role to a digital product.

        Parameters
        ----------
        digital_product_guid : str
            The GUID of the digital product.
        product_manager_role_guid : str
            The GUID of the product manager role.
        body : dict | NewRelationshipRequestBody, optional
            Request body containing relationship properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.

        Notes
        -----
        Sample JSON body:
        ```json
        {
          "class": "NewRelationshipRequestBody",
          "properties": {
              "class" : "AssignmentScopeProperties",
              "assignmentType": "Product Manager",
              "description": "The person or role responsible for the product"
          }
        }
        ```
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_link_product_manager(
                digital_product_guid, product_manager_role_guid, body
            )
        )

    @dynamic_catch
    async def _async_detach_product_manager(
        self,
        digital_product_guid: str,
        product_manager_role_guid: str,
        body: Optional[dict | DeleteRelationshipRequestBody] = None,
    ) -> None:
        """Detach a product manager from a digital product. Async version.

        Parameters
        ----------
        digital_product_guid : str
            The GUID of the digital product.
        product_manager_role_guid : str
            The GUID of the product manager role to detach.
        body : dict | DeleteRelationshipRequestBody, optional
            Request body for deletion.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.

        Notes
        -----
        Sample JSON body:
        ```json
        {
          "class" : "DeleteRelationshipRequestBody"
        }
        ```
        """
        url = (
            f"{self.product_manager_command_root}/digital-products/"
            f"{digital_product_guid}/product-managers/{product_manager_role_guid}/detach"
        )
        await self._async_delete_relationship_request(url, body)
        logger.info(f"Detached digital product manager {digital_product_guid} -> {product_manager_role_guid}")

    def detach_product_manager(
        self,
        digital_product_guid: str,
        product_manager_role_guid: str,
        body: Optional[dict | DeleteRelationshipRequestBody] = None,
    ) -> None:
        """Detach a product manager from a digital product.

        Parameters
        ----------
        digital_product_guid : str
            The GUID of the digital product.
        product_manager_role_guid : str
            The GUID of the product manager role to detach.
        body : dict | DeleteRelationshipRequestBody, optional
            Request body for deletion.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.

        Notes
        -----
        Sample JSON body:
        ```json
        {
          "class" : "DeleteRelationshipRequestBody"
        }
        ```
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_detach_product_manager(
                digital_product_guid, product_manager_role_guid, body
            )
        )

    #
    # Digital Product Catalog Management
    #

    @dynamic_catch
    async def _async_create_digital_product_catalog(
        self,
        body: Optional[dict | NewElementRequestBody] = None,
    ) -> str:
        """Create a new digital product catalog collection. Async version.

        Parameters
        ----------
        body : dict | NewElementRequestBody, optional
            Request body containing digital product catalog properties.

        Returns
        -------
        str
            The GUID of the created digital product catalog.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        ValidationError
            If the body does not conform to NewElementRequestBody.
        PyegeriaNotAuthorizedException
            If the user is not authorized for the requested action.

        Notes
        -----
            JSON Structure looks like:
            {
              "class" : "NewElementRequestBody",
              "typeName": "DigitalProductCatalog",
              "isOwnAnchor" : true,
              "anchorScopeGUID" : "optional GUID of search scope",
              "parentGUID" : "xxx",
              "parentRelationshipTypeName" : "CollectionMembership",
              "parentAtEnd1": true,
              "properties": {
                "class" : "CatalogProperties",
                "qualifiedName": "DigitalProductCatalog::Add catalog name here",
                "displayName" : "Catalog name",
                "description" : "Add description of catalog here",
                "additionalProperties": {
                  "property1Name" : "property1Value",
                  "property2Name" : "property2Value"
                }
              },
              "externalSourceGUID": "add guid here",
              "externalSourceName": "add qualified name here",
              "effectiveTime" : "timestamp",
              "forLineage" : false,
              "forDuplicateProcessing" : false,
            }
        """
        url = f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/product-manager/collections"
        return await self._async_create_element_body_request(url, ["CatalogProperties"], body)

    def create_digital_product_catalog(
        self,
        body: Optional[dict | NewElementRequestBody] = None,
    ) -> str:
        """Create a new digital product catalog collection. Sync version.

        Parameters
        ----------
        body : dict | NewElementRequestBody, optional
            Request body containing digital product catalog properties.

        Returns
        -------
        str
            The GUID of the created digital product catalog.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        body = self._prepare_body(body)
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self._async_create_digital_product_catalog(body))

    @dynamic_catch
    async def _async_update_digital_product_catalog(
        self,
        digital_product_catalog_guid: str,
        body: Optional[dict | UpdateElementRequestBody] = None,
    ) -> None:
        """Update the properties of a digital product catalog. Async version.

        Parameters
        ----------
        digital_product_catalog_guid : str
            The GUID of the digital product catalog to update.
        body : dict | UpdateElementRequestBody, optional
            Request body containing updated properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        ValidationError
            If the body does not conform to UpdateElementRequestBody.
        PyegeriaNotAuthorizedException
            If the user is not authorized for the requested action.
        """
        url = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/"
            f"product-manager/collections/{digital_product_catalog_guid}/update"
        )
        await self._async_update_element_body_request(url, ["CatalogProperties"], body)

    def update_digital_product_catalog(
        self,
        digital_product_catalog_guid: str,
        body: Optional[dict | UpdateElementRequestBody] = None,
    ) -> None:
        """Update the properties of a digital product catalog. Sync version.

        Parameters
        ----------
        digital_product_catalog_guid : str
            The GUID of the digital product catalog to update.
        body : dict | UpdateElementRequestBody, optional
            Request body containing updated properties.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        digital_product_catalog_guid = str(digital_product_catalog_guid)
        body = self._prepare_body(body)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_update_digital_product_catalog(digital_product_catalog_guid, body)
        )

    @dynamic_catch
    async def _async_delete_digital_product_catalog(
        self,
        digital_product_catalog_guid: str,
        body: Optional[dict | DeleteElementRequestBody] = None,
        cascade: bool = False,
    ) -> None:
        """Delete a digital product catalog. Async version.

        Parameters
        ----------
        digital_product_catalog_guid : str
            The GUID of the digital product catalog to delete.
        body : dict | DeleteElementRequestBody, optional
            Request body for deletion.
        cascade : bool, optional
            Whether to cascade the delete. Defaults to False.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        ValidationError
            If the body does not conform to DeleteElementRequestBody.
        PyegeriaNotAuthorizedException
            If the user is not authorized for the requested action.
        """
        url = (
            f"{self.platform_url}/servers/{self.view_server}/api/open-metadata/"
            f"product-manager/collections/{digital_product_catalog_guid}/delete"
        )
        await self._async_delete_element_request(url, body, cascade)

    def delete_digital_product_catalog(
        self,
        digital_product_catalog_guid: str,
        body: Optional[dict | DeleteElementRequestBody] = None,
        cascade: bool = False,
    ) -> None:
        """Delete a digital product catalog. Sync version.

        Parameters
        ----------
        digital_product_catalog_guid : str
            The GUID of the digital product catalog to delete.
        body : dict | DeleteElementRequestBody, optional
            Request body for deletion.
        cascade : bool, optional
            Whether to cascade the delete. Defaults to False.

        Returns
        -------
        None

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        digital_product_catalog_guid = str(digital_product_catalog_guid)
        body = self._prepare_body(body)
        cascade = bool(cascade)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            self._async_delete_digital_product_catalog(digital_product_catalog_guid, body, cascade)
        )

    @dynamic_catch
    async def _async_get_digital_product_catalog_by_guid(
        self,
        digital_product_catalog_guid: str,
        body: Optional[dict] = None,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
    ) -> dict | str:
        """Return the properties of a specific digital product catalog by GUID. Async version.

        Parameters
        ----------
        digital_product_catalog_guid : str
            The GUID of the digital product catalog to retrieve.
        body : dict, optional
            Request body (typically empty for retrieval).
        output_format : str, optional
            Format for output. Defaults to "JSON".
        report_spec : str | dict, optional
            Report specification for formatting.

        Returns
        -------
        dict | str
            The digital product catalog properties.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        PyegeriaNotFoundException
            If the digital product catalog is not found.
        PyegeriaNotAuthorizedException
            If the user is not authorized for the requested action.
        """
        url = f"{self.product_manager_command_root}/collections/{digital_product_catalog_guid}/retrieve"
        response = await self._async_get_guid_request(
            url,
            _type="DigitalProductCatalog",
            _gen_output=None,
            output_format=output_format,
            report_spec=report_spec,
            body=body,
        )
        return response

    def get_digital_product_catalog_by_guid(
        self,
        digital_product_catalog_guid: str,
        body: Optional[dict] = None,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
    ) -> dict | str:
        """Return the properties of a specific digital product catalog by GUID. Sync version.

        Parameters
        ----------
        digital_product_catalog_guid : str
            The GUID of the digital product catalog to retrieve.
        body : dict, optional
            Request body (typically empty for retrieval).
        output_format : str, optional
            Format for output. Defaults to "JSON".
        report_spec : str | dict, optional
            Report specification for formatting.

        Returns
        -------
        dict | str
            The digital product catalog properties.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self._async_get_digital_product_catalog_by_guid(
                digital_product_catalog_guid, body, output_format, report_spec
            )
        )

    @dynamic_catch
    async def _async_get_digital_product_catalogs_by_name(
        self,
        filter_string: Optional[str] = None,
        classification_names: Optional[list[str]] = None,
        body: Optional[dict] = None,
        start_from: int = 0,
        page_size: int = 0,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
    ) -> list | str:
        """Returns the list of digital product catalogs with a particular name. Async version.

        Parameters
        ----------
        filter_string : str, optional
            Filter string to match against catalog names.
        classification_names : list[str], optional
            List of classification names to filter by.
        body : dict, optional
            Request body for additional filtering.
        start_from : int, optional
            Starting index for pagination. Defaults to 0.
        page_size : int, optional
            Number of results per page. Defaults to 0 (no limit).
        output_format : str, optional
            Format for output. Defaults to "JSON".
        report_spec : str | dict, optional
            Report specification for formatting.

        Returns
        -------
        list | str
            List of digital product catalogs matching the criteria.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        url = f"{self.product_manager_command_root}/collections/by-name"
        response = await self._async_get_name_request(url, _type="DigitalProductCatalog", _gen_output=None,
                                                      filter_string=filter_string,
                                                      classification_names=classification_names, start_from=start_from,
                                                      page_size=page_size, output_format=output_format,
                                                      report_spec=report_spec, body=body)
        return response

    def get_digital_product_catalogs_by_name(
        self,
        filter_string: Optional[str] = None,
        classification_names: Optional[list[str]] = None,
        body: Optional[dict] = None,
        start_from: int = 0,
        page_size: int = 0,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
    ) -> list | str:
        """Returns the list of digital product catalogs with a particular name. Sync version.

        Parameters
        ----------
        filter_string : str, optional
            Filter string to match against catalog names.
        classification_names : list[str], optional
            List of classification names to filter by.
        body : dict, optional
            Request body for additional filtering.
        start_from : int, optional
            Starting index for pagination. Defaults to 0.
        page_size : int, optional
            Number of results per page. Defaults to 0 (no limit).
        output_format : str, optional
            Format for output. Defaults to "JSON".
        report_spec : str | dict, optional
            Report specification for formatting.

        Returns
        -------
        list | str
            List of digital product catalogs matching the criteria.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self._async_get_digital_product_catalogs_by_name(
                filter_string, classification_names, body, start_from, page_size, output_format, report_spec
            )
        )

    @dynamic_catch
    async def _async_find_digital_product_catalogs(
        self,
        search_string: str = "*",
        body: Optional[dict] = None,
        starts_with: bool = False,
        ends_with: bool = False,
        ignore_case: bool = True,
        start_from: int = 0,
        page_size: int = 0,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
        **kwargs
    ) -> list | str:
        """Returns the list of digital product catalogs matching the search string. Async version.

        Parameters
        ----------
        search_string : str, optional
            Search string to match. Defaults to "*" (all).
        starts_with : bool, optional
            Whether to match from the start. Defaults to False.
        ends_with : bool, optional
            Whether to match at the end. Defaults to False.
        ignore_case : bool, optional
            Whether to ignore case. Defaults to True.
        anchor_domain : str, optional
            Domain to anchor the search.
        metadata_element_type : str, optional
            Type of metadata element to search for.
        metadata_element_subtype : str, optional
            Subtype of metadata element.
        skip_relationships : list[str], optional
            Relationships to skip in the graph.
        include_only_relationships : list[str], optional
            Only include these relationships.
        skip_classified_elements : list[str], optional
            Skip elements with these classifications.
        include_only_classified_elements : list[str], optional
            Only include elements with these classifications.
        graph_query_depth : int, optional
            Depth of graph query. Defaults to 0.
        governance_zone_filter : list[str], optional
            Filter by governance zones.
        as_of_time : str, optional
            Historical time for the query.
        effective_time : str, optional
            Effective time for the query.
        relationship_page_size : int, optional
            Page size for relationships. Defaults to 0.
        limit_results_by_status : list[str], optional
            Limit results by status values.
        sequencing_order : str, optional
            Order for sequencing results.
        sequencing_property : str, optional
            Property to sequence by.
        start_from : int, optional
            Starting index for pagination. Defaults to 0.
        page_size : int, optional
            Number of results per page. Defaults to 0 (no limit).
        output_format : str, optional
            Format for output. Defaults to "JSON".
        report_spec : str | dict, optional
            Report specification for formatting.
        property_names: list[str], optional
            The names of properties to search for.
        body : dict, optional
            Request body for additional parameters.

        Returns
        -------
        list | str
            List of digital product catalogs matching the search criteria.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        url = f"{self.product_manager_command_root}/collections/by-search-string"
        
        # Merge explicit parameters with kwargs
        params = {
            'search_string': search_string,
            'body': body,
            'starts_with': starts_with,
            'ends_with': ends_with,
            'ignore_case': ignore_case,
            'start_from': start_from,
            'page_size': page_size,
            'output_format': output_format,
            'report_spec': report_spec
        }
        params.update(kwargs)
        
        # Filter out None values, but keep search_string even if None (it's required)
        params = {k: v for k, v in params.items() if v is not None or k == 'search_string'}
        
        response = await self._async_find_request(url, _type="DigitalProductCatalog", _gen_output=None, **params)
        return response

    def find_digital_product_catalogs(
        self,
        search_string: str = "*",
        body: Optional[dict] = None,
        starts_with: bool = False,
        ends_with: bool = False,
        ignore_case: bool = True,
        start_from: int = 0,
        page_size: int = 0,
        output_format: str = "JSON",
        report_spec: Optional[str | dict] = None,
        **kwargs
    ) -> list | str:
        """Returns the list of digital product catalogs matching the search string. Sync version.

        Parameters
        ----------
        search_string : str, optional
            Search string to match. Defaults to "*" (all).
        starts_with : bool, optional
            Whether to match from the start. Defaults to False.
        ends_with : bool, optional
            Whether to match at the end. Defaults to False.
        ignore_case : bool, optional
            Whether to ignore case. Defaults to True.
        anchor_domain : str, optional
            Domain to anchor the search.
        metadata_element_type : str, optional
            Type of metadata element to search for.
        metadata_element_subtype : str, optional
            Subtype of metadata element.
        skip_relationships : list[str], optional
            Relationships to skip in the graph.
        include_only_relationships : list[str], optional
            Only include these relationships.
        skip_classified_elements : list[str], optional
            Skip elements with these classifications.
        include_only_classified_elements : list[str], optional
            Only include elements with these classifications.
        graph_query_depth : int, optional
            Depth of graph query. Defaults to 0.
        governance_zone_filter : list[str], optional
            Filter by governance zones.
        as_of_time : str, optional
            Historical time for the query.
        effective_time : str, optional
            Effective time for the query.
        relationship_page_size : int, optional
            Page size for relationships. Defaults to 0.
        limit_results_by_status : list[str], optional
            Limit results by status values.
        sequencing_order : str, optional
            Order for sequencing results.
        sequencing_property : str, optional
            Property to sequence by.
        start_from : int, optional
            Starting index for pagination. Defaults to 0.
        page_size : int, optional
            Number of results per page. Defaults to 0 (no limit).
        output_format : str, optional
            Format for output. Defaults to "JSON".
        report_spec : str | dict, optional
            Report specification for formatting.
        property_names: list[str], optional
            The names of properties to search for.
        body : dict, optional
            Request body for additional parameters.

        Returns
        -------
        list | str
            List of digital product catalogs matching the search criteria.

        Raises
        ------
        PyegeriaException
            If there are issues in communications, message format, or Egeria errors.
        """
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(
            self._async_find_digital_product_catalogs(
                search_string=search_string,
                body=body,
                starts_with=starts_with,
                ends_with=ends_with,
                ignore_case=ignore_case,
                start_from=start_from,
                page_size=page_size,
                output_format=output_format,
                report_spec=report_spec,
                **kwargs
            )
        )


#!/usr/bin/env python3
"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.

Test script for the FormatSetDict class in _output_format_models.py.
This script tests the ability to find format sets by either name or alias.
"""

import sys
import os

# Add the parent directory to the path so we can import the modules
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from pyegeria.view._output_format_models import Column, Format, FormatSet, FormatSetDict

def test_format_set_dict_find_by_name_or_alias():
    """Test the ability to find format sets by either name or alias."""
    print("\nTesting FormatSetDict find_by_name_or_alias method...")
    
    # Create some test format sets with aliases
    format_set1 = FormatSet(
        heading="Test Format Set 1",
        description="A test format set with aliases",
        aliases=["test1", "format1", "set1"],
        formats=[
            Format(
                types=["TABLE"],
                columns=[
                    Column(name="Column 1", key="column1"),
                    Column(name="Column 2", key="column2"),
                ],
            ),
        ],
    )
    
    format_set2 = FormatSet(
        heading="Test Format Set 2",
        description="Another test format set with aliases",
        aliases=["test2", "format2", "set2"],
        formats=[
            Format(
                types=["DICT"],
                columns=[
                    Column(name="Column A", key="columnA"),
                    Column(name="Column B", key="columnB"),
                ],
            ),
        ],
    )
    
    # Create a FormatSetDict with the test format sets
    format_set_dict = FormatSetDict({
        "TestFormatSet1": format_set1,
        "TestFormatSet2": format_set2,
    })
    
    # Test find_by_name_or_alias method
    print("Testing find_by_name_or_alias method...")
    
    # Test finding by name
    result1 = format_set_dict.find_by_name_or_alias("TestFormatSet1")
    assert result1 is not None, "Expected to find format set by name"
    assert result1.heading == "Test Format Set 1", f"Expected 'Test Format Set 1', got '{result1.heading}'"
    print(f"Found format set by name: {result1.heading}")
    
    # Test finding by alias
    result2 = format_set_dict.find_by_name_or_alias("test1")
    assert result2 is not None, "Expected to find format set by alias"
    assert result2.heading == "Test Format Set 1", f"Expected 'Test Format Set 1', got '{result2.heading}'"
    print(f"Found format set by alias: {result2.heading}")
    
    # Test finding by another alias
    result3 = format_set_dict.find_by_name_or_alias("format2")
    assert result3 is not None, "Expected to find format set by alias"
    assert result3.heading == "Test Format Set 2", f"Expected 'Test Format Set 2', got '{result3.heading}'"
    print(f"Found format set by alias: {result3.heading}")
    
    # Test finding a non-existent format set
    result4 = format_set_dict.find_by_name_or_alias("NonExistentFormatSet")
    assert result4 is None, "Expected None for non-existent format set"
    print("Correctly returned None for non-existent format set")
    
    print("find_by_name_or_alias method test passed!")

def test_format_set_dict_get():
    """Test the get method with both names and aliases."""
    print("\nTesting FormatSetDict get method...")
    
    # Create some test format sets with aliases
    format_set1 = FormatSet(
        heading="Test Format Set 1",
        description="A test format set with aliases",
        aliases=["test1", "format1", "set1"],
        formats=[
            Format(
                types=["TABLE"],
                columns=[
                    Column(name="Column 1", key="column1"),
                    Column(name="Column 2", key="column2"),
                ],
            ),
        ],
    )
    
    format_set2 = FormatSet(
        heading="Test Format Set 2",
        description="Another test format set with aliases",
        aliases=["test2", "format2", "set2"],
        formats=[
            Format(
                types=["DICT"],
                columns=[
                    Column(name="Column A", key="columnA"),
                    Column(name="Column B", key="columnB"),
                ],
            ),
        ],
    )
    
    # Create a FormatSetDict with the test format sets
    format_set_dict = FormatSetDict({
        "TestFormatSet1": format_set1,
        "TestFormatSet2": format_set2,
    })
    
    # Test get method
    print("Testing get method...")
    
    # Test getting by name
    result1 = format_set_dict.get("TestFormatSet1")
    assert result1 is not None, "Expected to get format set by name"
    assert result1.heading == "Test Format Set 1", f"Expected 'Test Format Set 1', got '{result1.heading}'"
    print(f"Got format set by name: {result1.heading}")
    
    # Test getting by alias
    result2 = format_set_dict.get("test1")
    assert result2 is not None, "Expected to get format set by alias"
    assert result2.heading == "Test Format Set 1", f"Expected 'Test Format Set 1', got '{result2.heading}'"
    print(f"Got format set by alias: {result2.heading}")
    
    # Test getting by another alias
    result3 = format_set_dict.get("format2")
    assert result3 is not None, "Expected to get format set by alias"
    assert result3.heading == "Test Format Set 2", f"Expected 'Test Format Set 2', got '{result3.heading}'"
    print(f"Got format set by alias: {result3.heading}")
    
    # Test getting a non-existent format set
    result4 = format_set_dict.get("NonExistentFormatSet")
    assert result4 is None, "Expected None for non-existent format set"
    print("Correctly returned None for non-existent format set")
    
    # Test getting a non-existent format set with a default value
    default_value = "Default Value"
    result5 = format_set_dict.get("NonExistentFormatSet", default_value)
    assert result5 == default_value, f"Expected '{default_value}', got '{result5}'"
    print(f"Correctly returned default value for non-existent format set: {result5}")
    
    print("get method test passed!")

def test_format_set_dict_getitem():
    """Test the __getitem__ method with both names and aliases."""
    print("\nTesting FormatSetDict __getitem__ method...")
    
    # Create some test format sets with aliases
    format_set1 = FormatSet(
        heading="Test Format Set 1",
        description="A test format set with aliases",
        aliases=["test1", "format1", "set1"],
        formats=[
            Format(
                types=["TABLE"],
                columns=[
                    Column(name="Column 1", key="column1"),
                    Column(name="Column 2", key="column2"),
                ],
            ),
        ],
    )
    
    format_set2 = FormatSet(
        heading="Test Format Set 2",
        description="Another test format set with aliases",
        aliases=["test2", "format2", "set2"],
        formats=[
            Format(
                types=["DICT"],
                columns=[
                    Column(name="Column A", key="columnA"),
                    Column(name="Column B", key="columnB"),
                ],
            ),
        ],
    )
    
    # Create a FormatSetDict with the test format sets
    format_set_dict = FormatSetDict({
        "TestFormatSet1": format_set1,
        "TestFormatSet2": format_set2,
    })
    
    # Test __getitem__ method
    print("Testing __getitem__ method...")
    
    # Test getting by name
    try:
        result1 = format_set_dict["TestFormatSet1"]
        assert result1.heading == "Test Format Set 1", f"Expected 'Test Format Set 1', got '{result1.heading}'"
        print(f"Got format set by name: {result1.heading}")
    except KeyError:
        assert False, "Unexpected KeyError when getting by name"
    
    # Test getting by alias
    try:
        result2 = format_set_dict["test1"]
        assert result2.heading == "Test Format Set 1", f"Expected 'Test Format Set 1', got '{result2.heading}'"
        print(f"Got format set by alias: {result2.heading}")
    except KeyError:
        assert False, "Unexpected KeyError when getting by alias"
    
    # Test getting by another alias
    try:
        result3 = format_set_dict["format2"]
        assert result3.heading == "Test Format Set 2", f"Expected 'Test Format Set 2', got '{result3.heading}'"
        print(f"Got format set by alias: {result3.heading}")
    except KeyError:
        assert False, "Unexpected KeyError when getting by alias"
    
    # Test getting a non-existent format set
    try:
        result4 = format_set_dict["NonExistentFormatSet"]
        assert False, "Expected KeyError for non-existent format set"
    except KeyError:
        print("Correctly raised KeyError for non-existent format set")
    
    print("__getitem__ method test passed!")

def test_format_set_dict_contains():
    """Test the __contains__ method with both names and aliases."""
    print("\nTesting FormatSetDict __contains__ method...")
    
    # Create some test format sets with aliases
    format_set1 = FormatSet(
        heading="Test Format Set 1",
        description="A test format set with aliases",
        aliases=["test1", "format1", "set1"],
        formats=[
            Format(
                types=["TABLE"],
                columns=[
                    Column(name="Column 1", key="column1"),
                    Column(name="Column 2", key="column2"),
                ],
            ),
        ],
    )
    
    format_set2 = FormatSet(
        heading="Test Format Set 2",
        description="Another test format set with aliases",
        aliases=["test2", "format2", "set2"],
        formats=[
            Format(
                types=["DICT"],
                columns=[
                    Column(name="Column A", key="columnA"),
                    Column(name="Column B", key="columnB"),
                ],
            ),
        ],
    )
    
    # Create a FormatSetDict with the test format sets
    format_set_dict = FormatSetDict({
        "TestFormatSet1": format_set1,
        "TestFormatSet2": format_set2,
    })
    
    # Test __contains__ method
    print("Testing __contains__ method...")
    
    # Test checking by name
    assert "TestFormatSet1" in format_set_dict, "Expected format set to be found by name"
    print("Format set found by name")
    
    # Test checking by alias
    assert "test1" in format_set_dict, "Expected format set to be found by alias"
    print("Format set found by alias")
    
    # Test checking by another alias
    assert "format2" in format_set_dict, "Expected format set to be found by alias"
    print("Format set found by alias")
    
    # Test checking a non-existent format set
    assert "NonExistentFormatSet" not in format_set_dict, "Expected format set not to be found"
    print("Non-existent format set correctly not found")
    
    print("__contains__ method test passed!")

def main():
    """Run all tests."""
    test_format_set_dict_find_by_name_or_alias()
    test_format_set_dict_get()
    test_format_set_dict_getitem()
    test_format_set_dict_contains()
    
    print("\nAll tests passed!")

if __name__ == "__main__":
    main()


"""
SPDX-License-Identifier: Apache-2.0
Copyright Contributors to the ODPi Egeria project.



Execute Glossary actions.

"""

import os

import click


from pyegeria import (
    EgeriaTech, body_slimmer, settings,
    PyegeriaAPIException, PyegeriaClientException, print_basic_exception, print_exception_table
)
app_config = settings.Environment

EGERIA_USER = os.environ.get("EGERIA_USER", "erinoverview")
EGERIA_USER_PASSWORD = os.environ.get("EGERIA_USER_PASSWORD", "secret")

COLLECTION_TYPES = ["Collection", "Glossary", "CollectionFolder", "RootCollection", "WorkItemList", "SolutionBlueprint" ]


@click.command("add-element-to-collection")
@click.option("--element-guid", required=True, default = "",help="GUID of the element to add to a Collection")
@click.option("--collection-guid", required=True, default = "", help="GUID of the Collection to add an element to")
@click.option("--server", default=app_config.egeria_view_server, help="Egeria view server to use")
@click.option(
    "--url", default=app_config.egeria_view_server_url, help="URL of Egeria platform to connect to"
)
@click.option("--userid", default=EGERIA_USER, help="Egeria user")
@click.option("--password", default=EGERIA_USER_PASSWORD, help="Egeria user password")
@click.option("--timeout", default=60, help="Number of seconds to wait")


def add_element_to_collection(server, url, userid, password, timeout, element_guid, collection_guid):
    """Add a glossary term to a Collection"""
    m_client = EgeriaTech(server, url, user_id=userid, user_pwd=password)
    token = m_client.create_egeria_bearer_token()
    try:
        element_guid = element_guid.strip()
        collection_guid = collection_guid.strip()
        m_client.add_to_collection(element_guid, collection_guid)

        click.echo(
            f"Added element with GUID: {element_guid} to Collection with GUID: {collection_guid}\n"
        )

    except (PyegeriaAPIException, PyegeriaClientException) as e:
        print_basic_exception(e)
    finally:
        m_client.close_session()

@click.command("remove-element-from-collection")
@click.option("--element-guid", required=True, default = "",help="GUID of the element to remove from the collection")
@click.option("--collection-guid", required=True, default = "", help="GUID of collection to remove the element from")

@click.option("--server", default=app_config.egeria_view_server, help="Egeria view server to use")
@click.option(
    "--url", default=app_config.egeria_view_server_url, help="URL of Egeria platform to connect to"
)
@click.option("--userid", default=EGERIA_USER, help="Egeria user")
@click.option("--password", default=EGERIA_USER_PASSWORD, help="Egeria user password")
@click.option("--timeout", default=60, help="Number of seconds to wait")

def remove_element_from_collection(server, url, userid, password, timeout, element_guid, collection_guid):
    """Remove an element from a Collection"""
    m_client = EgeriaTech(server, url, user_id=userid, user_pwd=password)
    token = m_client.create_egeria_bearer_token()
    try:
        element_guid = element_guid.strip()
        collection_guid = collection_guid.strip()
        body = {
            "class": "DeleteRelationshipRequestBody"
        }
        m_client.remove_from_collection( collection_guid, element_guid, body)

        click.echo(
            f"Removed term with GUID: {element_guid} from Collection with GUID: {collection_guid}\n"
        )

    except (PyegeriaAPIException, PyegeriaClientException) as e:
        print_basic_exception(e)
    finally:
        m_client.close_session()


@click.command("create-collection")
@click.option("--name", help="Collection Name",default = "", required=True)
@click.option("--kind", help="Kind of Collection", default = "Collection",
            type = click.Choice(COLLECTION_TYPES, case_sensitive=False))
@click.option(
    "--description",
    help="Description of the Collection",
    default="A description goes here",
)
@click.option(
    "--category", default = "",
    help="Category of the Collection"
)
@click.option("--server", default=app_config.egeria_view_server, help="Egeria view server to use.")
@click.option(
    "--url", default=app_config.egeria_view_server_url, help="URL of Egeria platform to connect to"
)
@click.option("--userid", default=EGERIA_USER, help="Egeria user")
@click.option("--password", default=EGERIA_USER_PASSWORD, help="Egeria user password")
@click.option("--timeout", default=60, help="Number of seconds to wait")
def create_collection(
    server: str,
    url: str,
    userid: str,
    password: str,
    timeout: int,
    name: str,
    kind: str,
    description: str,
    category: str,
) -> None:
    """Create a new Collection"""

    try:
        m_client = EgeriaTech(server, url, userid, password)
        token = m_client.create_egeria_bearer_token()

        collection_guid = m_client.create_collection( display_name = name, description = description, category = category,
                                                      prop = [kind])
        print(f"New categry \'{name}\' created with id of \'{collection_guid}\'")

    except (PyegeriaAPIException, PyegeriaClientException) as e:
        print_basic_exception(e)
    finally:
        m_client.close_session()


@click.command("update-collection")
@click.option("--name", help="Collection Name", default = "",required=False)
@click.option("--description", default = "",help="Description of the Collection")
@click.option("--category", default = "", help="Category of the Collection")
@click.option("--server", default=app_config.egeria_view_server, help="Egeria view server to use")
@click.option(
    "--url", default=app_config.egeria_view_server_url, help="URL of Egeria platform to connect to"
)
@click.option("--userid", default=EGERIA_USER, help="Egeria user")
@click.option("--password", default=EGERIA_USER_PASSWORD, help="Egeria user password")
@click.option("--timeout", default=60, help="Number of seconds to wait")
@click.argument("collection-guid")

def update_collection(collection_guid, name, description, category, server, url, userid, password, timeout, ):
    """Update the collection specified"""
    m_client = EgeriaTech(server, url, user_id=userid, user_pwd=password)
    token = m_client.create_egeria_bearer_token()
    try:
        body = {
          "class" : "UpdateElementRequestBody",
          "properties": {
            "class" : "CollectionProperties",
            "name" : name,
            "description" : description,
            "category": category
          },
        }
        m_client.update_collection(collection_guid, body_slimmer(body))

        click.echo(f"Updated glossary: {collection_guid}")

    except (PyegeriaAPIException, PyegeriaClientException) as e:
        print_basic_exception(e)
    finally:
        m_client.close_session()

@click.command("delete-collection")
@click.argument("collection-guid")
@click.option("--server", default=app_config.egeria_view_server, help="Egeria view server to use")
@click.option(
    "--url", default=app_config.egeria_view_server_url, help="URL of Egeria platform to connect to"
)
@click.option("--userid", default=EGERIA_USER, help="Egeria user")
@click.option("--password", default=EGERIA_USER_PASSWORD, help="Egeria user password")
@click.option("--timeout", default=60, help="Number of seconds to wait")

def delete_collection(server, url, userid, password, timeout, collection_guid):
    """Delete the collection specified"""
    m_client = EgeriaTech(server, url, user_id=userid, user_pwd=password)
    token = m_client.create_egeria_bearer_token()
    try:
        m_client.delete_collection(collection_guid)

        click.echo(f"Deleted collection: {collection_guid}")

    except (PyegeriaAPIException, PyegeriaClientException) as e:
        print_basic_exception(e)
    finally:
        m_client.close_session()